part of model_notifier;

/// Class for sending queries.
///
/// Basically, it allows you to send the same query as Firestore.
///
/// Can be converted to a [String] by passing [value].
@immutable
class Query {
  /// Class for sending queries.
  ///
  /// Basically, it allows you to send the same query as Firestore.
  ///
  /// Can be converted to a [String] by passing [value].
  const Query(
    this.path, {
    this.isEqualTo,
    this.isNotEqualTo,
    // this.isLessThan,
    this.isLessThanOrEqualTo,
    // this.isGreaterThan,
    this.isGreaterThanOrEqualTo,
    this.arrayContains,
    this.arrayContainsAny,
    this.whereIn,
    this.whereNotIn,
  });

  /// Class for sending queries.
  ///
  /// Basically, it allows you to send the same query as Firestore.
  ///
  /// Can be converted to a [String] by passing [value].
  factory Query.fromPath(String path) {
    if (path.isEmpty) {
      return Query(path);
    }
    final uri = Uri.tryParse(path);
    if (uri == null) {
      return Query(path);
    }
    final query = uri.queryParameters;

    return Query(
      uri.path,
      isEqualTo: _parseQuery(query, "equalTo"),
      isNotEqualTo: _parseQuery(query, "notEqualTo"),
      isLessThanOrEqualTo: _parseQuery(query, "endAt"),
      isGreaterThanOrEqualTo: _parseQuery(query, "startAt"),
      arrayContains: _parseQuery(query, "contains"),
      arrayContainsAny: _parseQuery(query, "containsAny", true),
      whereIn: _parseQuery(query, "whereIn", true),
      whereNotIn: _parseQuery(query, "whereNotIn", true),
    );
  }

  static dynamic _parseQuery(Map<String, String> query, String key,
      [bool isArray = false]) {
    if (!query.containsKey(key)) {
      return null;
    }
    final value = query[key];
    if (value.isEmpty) {
      return null;
    }
    if (isArray) {
      return value!.split(",").mapAndRemoveEmpty((item) => item.toAny());
    }
    return value.toAny();
  }

  /// Query path.
  final String path;

  /// If the value of [key] is equal to this value, `true`.
  final dynamic isEqualTo;

  /// If the value of [key] is not equal to this value, `true`.
  final dynamic isNotEqualTo;
  // final dynamic isLessThan;
  /// If the value of [key] is less than this value, `true`.
  final dynamic isLessThanOrEqualTo;
  // final dynamic isGreaterThan;
  /// If the value of [key] is greater than this value, `true`.
  final dynamic isGreaterThanOrEqualTo;

  /// If this value is in the [key] array, `true`.
  final dynamic arrayContains;

  /// If the [key] array contains one of these values, `true`.
  final DynamicList? arrayContainsAny;

  /// If the value of [key] is equal to one of these values, `true`.
  final DynamicList? whereIn;

  /// If the value of [key] is not equal to all of these values, `true`.
  final DynamicList? whereNotIn;

  /// Convert all Query to [String] parameters.
  String get value {
    assert(
      isEqualTo == null &&
          isNotEqualTo == null &&
          isLessThanOrEqualTo == null &&
          isGreaterThanOrEqualTo == null &&
          arrayContains == null &&
          arrayContainsAny == null &&
          whereIn == null &&
          whereNotIn == null,
      "If you want to specify a condition, please specify [key].",
    );
    if (isEqualTo != null) {
      return "$path?${"&equalTo=$isEqualTo".trimString("&")}";
    } else if (isNotEqualTo != null) {
      return "$path?${"&notEqualTo=$isNotEqualTo".trimString("&")}";
    } else if (isLessThanOrEqualTo != null) {
      return "$path?${"&endAt=$isLessThanOrEqualTo".trimString("&")}";
    } else if (isGreaterThanOrEqualTo != null) {
      return "$path?${"&startAt=$isGreaterThanOrEqualTo".trimString("&")}";
    } else if (arrayContains != null) {
      return "$path?${"&contains=$arrayContains".trimString("&")}";
    } else if (arrayContainsAny != null) {
      return "$path?" +
          "&containsAny=${arrayContainsAny!.map((e) => e.toString()).join(",")}"
              .trimString("&");
    } else if (whereIn != null) {
      return "$path?" +
          "&whereIn=${whereIn!.map((e) => e.toString()).join(",")}"
              .trimString("&");
    } else if (whereNotIn != null) {
      return "$path?" +
          "&whereNotIn=${whereNotIn!.map((e) => e.toString()).join(",")}"
              .trimString("&");
    }
    return path;
  }

  /// Check if the condition is satisfied by passing [data].
  ///
  /// When [DynamicMap] is passed, the value for [key] is checked.
  ///
  /// If `[List]' is passed,
  /// it checks whether or not the elements of the list satisfy the condition,
  /// and returns `true` if one of them satisfies the condition.
  ///
  /// All other values are checked to see if they satisfy the condition themselves, regardless of [key].
  bool check(Object? data) {
    if (data == null) {
      return false;
    }
    if (data is List) {
      return data.any((element) => check(element));
    } else {
      return _check(data);
    }
  }

  bool _check(dynamic value) {
    if (isEqualTo != null) {
      return value == isEqualTo;
    }
    if (isNotEqualTo != null) {
      return value != isNotEqualTo;
    }
    if (isGreaterThanOrEqualTo != null) {
      return value >= isGreaterThanOrEqualTo;
    }
    if (isLessThanOrEqualTo != null) {
      return value <= isLessThanOrEqualTo;
    }
    if (arrayContains != null) {
      final list = value;
      if (list is! List) {
        return false;
      }
      return list.contains(arrayContains);
    }
    if (arrayContainsAny.isNotEmpty) {
      final list = value;
      if (list is! List) {
        return false;
      }
      final any = arrayContainsAny!.mapAndRemoveEmpty((item) => item.toAny());
      return list.any((e) => e == any);
    }
    if (whereIn.isNotEmpty) {
      final val = value;
      final any = whereIn!.mapAndRemoveEmpty((item) => item.toAny());
      return any.contains(val);
    }
    if (whereNotIn.isNotEmpty) {
      final val = value;
      final any = whereNotIn!.mapAndRemoveEmpty((item) => item.toAny());
      return !any.contains(val);
    }
    return true;
  }

  /// Pass [list], which is a List of DynamicMap,
  /// to filter only the elements that satisfy the condition.
  List<T> filter<T>(List<T> list) {
    return list.where((element) => check(element)).toList();
  }

  /// Convert the query from [DynamicMap].
  static Query? _fromMap(DynamicMap map) {
    if (map.isEmpty || !map.containsKey("path")) {
      return null;
    }

    return Query(
      map.get("path", ""),
      isEqualTo: map.get<dynamic>("equalTo", null),
      isNotEqualTo: map.get<dynamic>("notEqualTo", null),
      isLessThanOrEqualTo: map.get<dynamic>("endAt", null),
      isGreaterThanOrEqualTo: map.get<dynamic>("startAt", null),
      arrayContains: map.get<dynamic>("contains", null),
      arrayContainsAny: map.get<List?>("containsAny", null),
      whereIn: map.get<List?>("whereIn", null),
      whereNotIn: map.get<List?>("whereNotIn", null),
    );
  }

  /// Convert the query to [DynamicMap].
  DynamicMap toMap() {
    return <String, dynamic>{
      "path": path,
      if (isEqualTo != null) "equalTo": isEqualTo,
      if (isNotEqualTo != null) "notEqualTo": isNotEqualTo,
      if (isLessThanOrEqualTo != null) "endAt": isLessThanOrEqualTo,
      if (isGreaterThanOrEqualTo != null) "startAt": isGreaterThanOrEqualTo,
      if (arrayContains != null) "contains": arrayContains,
      if (arrayContainsAny.isNotEmpty) "containsAny": arrayContainsAny,
      if (whereIn.isNotEmpty) "whereIn": whereIn,
      if (whereNotIn.isNotEmpty) "whereNotIn": whereNotIn,
    };
  }
}

extension CQueryDynamicMapExtensions on DynamicMap {
  /// Convert the query from [DynamicMap].
  Query? toQuery() {
    return Query._fromMap(this);
  }
}
